/* includes *******************************************************************/

#include <Tree.h>

#include <w3c/svg/SVGTransforms.hpp>
#include <w3c/svg/SVGTransformable.hpp>

#include <hsm/svgl/Svgl.h>


/* ANode implementation *******************************************************/

ANode::ANode(svg::SVGElement *e, hsm::SVGLWindow *w) : hsm::SVGLBehaviour(e, w) {
	// retriving parent
	hsm::svgl::getAttribute(elem, "hsm-arg:parent", parent);
}

unsigned int ANode::getHeight() {
	return 1;
}

void ANode::translateAllBrothers(int dy) {
	translateBrothers(dy);
	if((parent != 0) && (parent->isOpen)) {
		parent->translateAllBrothers(dy);
	}
}

void ANode::translateBrothers(int dy) {
	dom::Node *sibling = elem->getNextSibling();
	svg::SVGTransformable *transSibling;
	while((transSibling = hsm::svgl::getTypedBrother< svg::SVGTransformable * >(sibling)) != 0) {
		svg::SVGTranslateTransform *trans = transSibling->getTranslate(true);
		trans->ty += dy * 16;
	}
	if(parent != 0) parent->resizeBackground(dy);
}


/* Leaf class implementation **************************************************/

Leaf::Leaf(svg::SVGElement *e, hsm::SVGLWindow *w) : ANode(e, w) {
	translateAllBrothers(1);
}

Leaf::~Leaf() {
	window->lock();
	translateAllBrothers(-1);
	if(parent != 0) {
		parent->remove(elem);
	}
	window->unlock();
}

const hsm::SVGLBehaviour::Factory::Register< Leaf > Leaf::registration("treeLeaf");


/* Node class implementation **************************************************/

Node::Node(svg::SVGElement *e, hsm::SVGLWindow *w) : ANode(e, w) {
	// retriving the various svg parts
	dom::Node *child = elem->getFirstChild();
	background = hsm::svgl::getTypedBrother< svg::SVGRectElement * >(child);
	handle = hsm::svgl::getTypedBrother< svg::SVGUseElement * >(child);
	content = hsm::svgl::getTypedBrother< svg::SVGGElement * >(child);

	// initializing content
	child = content->getFirstChild();
	svg::SVGGElement *gChild;
	while((gChild = hsm::svgl::getTypedBrother< svg::SVGGElement * >(child)) != 0) {
		hsm::svgl::setAttribute(gChild, "hsm-arg:parent", this);
	}
	window->instanciateBehaviour(content);

	// initializing handle
	hsm::svgl::setAttribute(handle, "hsm:behaviour", "treeOpener");
	hsm::svgl::setAttribute(handle, "hsm-arg:parent", this);
	window->instanciateBehaviour(handle);

	unsigned int height = getHeight();
	translateAllBrothers(height);
	resizeBackground(height - 1);
}

Node::~Node() {
	window->lock();
	translateAllBrothers(-getHeight());
	if(parent != 0) {
		parent->remove(elem);
	}
	window->unlock();
}


void Node::open() {
	isOpen = true;
	
	// refreshing display
	window->lock();

	handle->setHref("#opened");
	elem->appendChild(content);

	unsigned int height = getHeight();
	resizeBackground(height - 1);
	translateAllBrothers(height - 1);

	window->unlock();
}

void Node::close() {
	// updating display
	window->lock();
	
	handle->setHref("#closed");
	elem->removeChild(content);
	background->setHeight(16);

	translateAllBrothers(1 - getHeight());

	window->unlock();

	isOpen = false;
}

void Node::resizeBackground(int dHeight) {
	if(isOpen) {
		background->setHeight(background->getHeight().getValue() + dHeight * 16);
	}
}

unsigned int Node::getHeight() {
	unsigned int res = 1;
	if(isOpen) {
		dom::Node *child = content->getFirstChild();
		svg::SVGGElement *gChild;
		while((gChild = hsm::svgl::getTypedBrother< svg::SVGGElement * >(child)) != 0) {
			ANode *node = dynamic_cast< ANode * >(hsm::SVGLBehaviour::Factory::get(gChild));
			if(node != 0) {
				res += node->getHeight();
			}
		}
	}
	return res;
}

void Node::add(svg::SVGElement *e, unsigned int p) {
	window->lock();
	dom::Node *child = content->getFirstChild();
	svg::SVGGElement *gChild;
	unsigned int i = 0;
	unsigned int height = 0;
	while(((gChild = hsm::svgl::getTypedBrother< svg::SVGGElement * >(child)) != 0) && (i++ < p)) {
		ANode *node = dynamic_cast< ANode * >(hsm::SVGLBehaviour::Factory::get(gChild));
		if(node != 0) {
			height += node->getHeight();
		}
	}
	
	if(gChild == 0) {
		content->appendChild(e);
	} else {
		content->insertBefore(e, gChild);
	}
	
	svg::SVGTransformable *transformable = dynamic_cast< svg::SVGTransformable * >(e);
	svg::SVGTranslateTransform *trans = transformable->getTranslate(true);
	trans->ty += height * 16;

	hsm::svgl::setAttribute(e, "hsm-arg:parent", this);
	window->instanciateBehaviour(e);
	window->unlock();
}

void Node::remove(svg::SVGElement *e) {
	content->removeChild(e);
}

const hsm::SVGLBehaviour::Factory::Register< Node > Node::registration("treeNode");
