////////////////////////////////////////////////////////////////////////////////
// File    : Point.cpp
// Content : Point implementation
// History : 03-Dec-2001 - [rb] creation
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/* includes *******************************************************************/

#include <Point.h>

#include <hsm/util/Debug.h>

#include <sstream>
#include <cmath>


/* macros *********************************************************************/

#define MIN(n1, n2) ((n1 < n2)? n1 : n2)


namespace hsm {

/* Point class implementation *************************************************/

	const Point *Point::cast = 0;
	
	Point::Point(const Id &id, unsigned int d, float th /* = 1. */)
	 : impl::Component< Composite >(id), dimension(d), threshold2(th * th), diff2(0) {
		construct();
	}

	Point::Point(int d, float th /* = 1. */)
	 : impl::Component< Composite >(), dimension(d), threshold2(th * th), diff2(0) {
		construct();
	}

	void Point::construct() {
		values.resize(dimension);
		for(unsigned int i = 0; i < dimension; ++i) {
			std::stringstream valueId;
			valueId << i;
			values[i] = new Value(valueId.str());
			add(values[i]);
		}
	}

	Point::~Point() {
		for(unsigned int i = 0; i < dimension; ++i) {
			remove(values[i]);
			delete values[i];
		}
	}


	void Point::process(const Event &event) {
		dispatch(event.message, event.data);
	}


	Point &Point::operator=(float f) {
		current = 0;
		return this->operator,(f);
	}

	Point &Point::operator,(float f) {
		if(current < dimension) {
			Value *val = values[current++];
			diff2 += pow(f - *val, 2);
			val->set(f);
		}
		if((current == dimension) && (diff2 >= threshold2)) {
			diff2 = 0;
			dispatch();
		}
		return *this;
	}


	Point &Point::operator=(const Point &p) {
		for(unsigned int i = 0; i < MIN(dimension, p.dimension); ++i) {
			(*values[i]).set(*p.values[i]);
		}
		dispatch();
		return *this;
	}

	Point &Point::operator+=(const Point &p) {
		for(unsigned int i = 0; i < MIN(dimension, p.dimension); ++i) {
			(*values[i]).add(*p.values[i]);
		}
		dispatch();
		return *this;
	}

	Point &Point::operator-=(const Point &p) {
		for(unsigned int i = 0; i < MIN(dimension, p.dimension); ++i) {
			(*values[i]).sub(*p.values[i]);
		}
		dispatch();
		return *this;
	}


	Point &Point::operator*=(float f) {
		for(unsigned int i = 0; i < dimension; ++i) {
			(*values[i]).mul(f);
		}
		dispatch();
		return *this;
	}

	Point &Point::operator/=(float f) {
		for(unsigned int i = 0; i < dimension; ++i) {
			(*values[i]).div(f);
		}
		dispatch();
		return *this;
	}
	
	
	float Point::operator*(const Point &p) const {
		float l = 0;
		for(unsigned int i = 0; i < MIN(dimension, p.dimension); ++i) {
			l += (*values[i]) * (*p.values[i]);
		}
		return l;		
	}

	float abs2(const Point &p) {
		float l = 0;
		for(unsigned int i = 0; i < p.dimension; l += pow(*p.values[i++], 2));
		return l;
	}

	float abs(const Point &p) {
		return sqrt(abs2(p));
	}

	
	bool Point::operator<=(const Point &p) const {
		return (abs2(*this) <= abs2(p));
	}
	
	bool Point::operator>=(const Point &p) const {
		return (abs2(*this) >= abs2(p));
	}
	
	
	Value &operator/(const Component &point, Point::ValueId i) {
		const Point *p = dynamic_cast< const Point * >(&point);
		return *(p->values[i]);
	}

	std::ostream &operator<<(std::ostream &s, const Point &point) {
		s << "Point(";
		if(point.dimension > 0) {
			s << *point.values[0];
		}
		for(unsigned int i = 1; i < point.dimension; ++i) {
			s << "," << *point.values[i];
		}
		return s << ")";
	}


} // namespace hsm


/* macros *********************************************************************/

#undef MIN
