////////////////////////////////////////////////////////////////////////////////
// File    : Protocol.h
// Content : interaction protocols
// History : 25-Aug-2004 - [rb] creation
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/// \file Protocol.h
/// Basic abstract interaction protocols definitions


#if !defined(HSM_PROTOCOL_H)
#define HSM_PROTOCOL_H


/* includes *******************************************************************/

#include <hsm/component/Point.h>

	
namespace hsm {

/* Protocol struct definition *************************************************/

	/// Common abstract protocol interface
	struct Protocol {
	protected :
		Point pick; ///< where the object has been picked
		
	public :
	
		Protocol() : pick(2) {} ///< constructor
		virtual ~Protocol()  {} ///< destructor

		/// returns whether the caller can start interacting or not
		/// \return yes or no
		virtual bool begin() { return true; }
		virtual void end()   {} ///< ending an interaction

		/// set where the object has been picked
		void setPickPoint(const Point &p) { pick = p; }
	};
	

/* EnterLeave protocol declaration ********************************************/

	/// Things that can be entered and leaved
	struct EnterLeave : virtual Protocol {
		virtual void enter() {}; ///< what to do when entered
		virtual void leave() {}; ///< what to do when lived 
	};
	
	/// Code wrapper around EnterLeave protocol
	template< class Implementation >
	struct EnterLeaveWrapper : EnterLeave, virtual Implementation {
		virtual void enter() { Implementation::enter(); }
		virtual void leave() { Implementation::leave(); }
	};
	
	
/* PushPop protocol declaration ***********************************************/

	/// Things that can be pushed and poped
	struct PushPop : virtual Protocol {
		virtual void push() {}; ///< what if pushed ?
		virtual void pop()  {}; ///< and what if poped ?
	};
	
	/// Code wrapper around PushPop protocol
	template< class Implementation >
	struct PushWrapper : virtual PushPop, virtual Implementation {
		virtual void push() { Implementation::push(); }
	};

	/// Code wrapper around PushPop protocol
	template< class Implementation >
	struct PopWrapper : virtual PushPop, virtual Implementation {
		virtual void pop()  { Implementation::pop(); }
	};

	/// Code wrapper around PushPop protocol
	template< class Implementation >
	struct PushPopWrapper : PushWrapper< Implementation >, PopWrapper< Implementation > {
	};
	
	
/* Translate protocol declaration *********************************************/

	/// Things that can be Translated
	struct Translate : virtual Protocol {
		/// what to do when translated
		/// \param dp delta p
		virtual void translate(const Point &dp) {};	
	};
	
	/// Code wrapper around Translate protocol
	template< class Implementation >
	struct TranslateWrapper : Translate, virtual Implementation {
		virtual void translate(const Point &dp) { Implementation::translate(p); }
	};


/* Rotate protocol declaration ************************************************/

	/// Things that can be Rotated
	struct Rotate : virtual Protocol {
		/// what to do when translated
		/// \param da delta angle
		virtual void rotate(const Value&da) {};
	};
	
	/// Code wrapper around Rotate protocol
	template< class Implementation >
	struct RotateWrapper : Rotate, virtual Implementation {
		virtual void rotate(const Value &da) { Implementation::rotate(da); }
	};
	
	
/* Zoom protocol declaration **************************************************/

	struct Zoom : virtual Protocol {
		/// what to do when zoomed
		/// \param ds delta scale
		virtual void zoom(float ds = 1.) {};
	};
	
	/// Code wrapper around Zoom protocol
	template< class Implementation >
	struct ZoomWrapper : Zoom, virtual Implementation {
		virtual void zoom(float ds = 1.) { Implementation::zoom(ds); }
	};
	

} // namespace hsm


#endif // !defined(HSM_PROTOCOL_H)
