////////////////////////////////////////////////////////////////////////////////
// File    : Selector.h
// Content : A class for select operations
// History : 31-Jan-2003 - [rb] creation
//           08-Feb-2003 - [rb] clean-up and first functional version
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/// \file Selector.h
/// a class for watching selectable objects


#if !defined(HSM_SELECTOR_H)
#define HSM_SELECTOR_H


/* includes *******************************************************************/

#include <hsm/system/select/File.h>

#include <map>


namespace hsm {

/* forward declarations *******************************************************/
	
	class Alert;
	class Mutex;
	class ThreadLoop;
	class Functor;
	

/* type definitions ***********************************************************/

	
	
/* Selector class definition **************************************************/

	/// implements a \c select abstraction
	/// A Selector watches os file descriptors or Selectable objects and triggers
	/// actions when they are in a particular state.
	/// \see hsm::Selectable
	/// \see hsm::file::descriptor
	class Selector {

		struct os {
			static void watch(void *selector);
		};

	public :
		/// status used for watching file descriptors
		enum Status {
			N   = 0, ///< nothing to watch
			R   = 1, ///< ready for reading
			W   = 2, ///< ready for writing
			RW  = 3, ///< ready for reading & writing
			E   = 4, ///< ready for ?
			RWE = 7  ///< ready for reading, writing & ?
		};

	protected :
		typedef std::map< Functor *, Status > StatusMap;
		typedef std::map< file::Descriptor, StatusMap > StatusMapMap;

		Alert *alert;
		Mutex *mutex;
		
		Functor    *selectorFn;
		ThreadLoop *threadLoop;

		StatusMapMap statusMapMap;
		
	public :
		         Selector(); ///< constructor
		virtual ~Selector(); ///< destructor

		/// adds a os file descriptor to watch
		/// \param fn a pointer on the function to callback
		/// \param fd the file (or Selectable) to watch
		/// \param status the status that triggers the Selector
		void add(Functor *fn, file::Descriptor fd, Status status = RWE);
		
		/// removes a os file descriptor to watch
		/// \param fn a pointer on the function called back		
		/// \param fd the file (or Selectable) watched
		/// \param status the status that triggerd the Selector
		void remove(Functor *fn, file::Descriptor fd, Status status = RWE);
	};
	

} // namespace hsm


#endif // !defined(HSM_SELECTOR_H)
