////////////////////////////////////////////////////////////////////////////////
// File    : Window.WIN32.cpp
// Content : Window::os struct implementation
// History : 11-Oct-2001 - [rb] creation
//           01-Feb-2004 - [rb] put the window to foreground & hide the console
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/* includes *******************************************************************/

#include <Window.h>

#include <hsm/component/Point.h>
#include <hsm/system/window/EventHandler.h>

#define _WIN32_WINNT 0x0400
#include <windows.h>


namespace hsm {

/* static variable initialisation *********************************************/

	Window::os::Map Window::os::windowMap;


/* Window::os struct implementation *******************************************/

	LONG WINAPI Window::os::defaulttHandler(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam) {
		long res = 0;
		hsm::Window *hw = windowMap[hWnd];

		switch(iMessage) {
		case WM_CLOSE :
			exit(EXIT_SUCCESS);
			break;

		case WM_MOVE :
			*(hw->position) = rect.left, rect.top;
			break;
		case WM_SIZE :
			*(hw->size) = (rect.right - rect.left), (rect.bottom - rect.top);
			break;
		}

		// dispatching system messages
		for(Handlers::iterator i = window->handlers.begin(); (i != window->handlers.end()) && (res == 0); ++i) {
			res = (*i)(hWnd, iMessage, wParam, lParam);
		}

		if(res == 0) {
			res = DefWindowProc(hWnd, iMessage, wParam, lParam);
		}

		return res;
	}

			
	Window::os::Window Window::os::create(hsm::Window *hw, const std::string &title, int x, int y, int w, int h) {
		HINSTANCE hInstance = HINSTANCE(GetModuleHandle(0));
		const char *name = title.c_str();

		// window initialisation
		WNDCLASS wc;
		wc.style = CS_BYTEALIGNCLIENT;
		wc.lpfnWndProc = &DefWindowProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = hInstance;
		wc.hIcon = 0;
		wc.hCursor = LoadCursor(0, IDC_ARROW);
		wc.hbrBackground = HBRUSH(GetStockObject(WHITE_BRUSH));
		wc.lpszMenuName = name;
		wc.lpszClassName = name;
		RegisterClass(&wc);

		Window::os::Window window = CreateWindow(name, name,
		                                         WS_TILEDWINDOW,
		                                         x, y, w, h, // CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		                                         0, 0,
		                                         hInstance,
 		                                         0);
		
		windowMap[window] = hw;
		setHandler(window);
		
		return window;
	}
	
	void Window::os::destroy(Window window) {
		DestroyWindow(window);
	}

	void Window::os::setHandler(Window window) {
		SetWindowLong(window, GWL_WNDPROC, reinterpret_cast<long>(defaultHandler));
	}
	
	void Window::os::setPosition(Window window, int x, int y, int w, int h) {
		MoveWindow(window, x, y, w, h, true);
	}
	
	void Window::os::setSize(Window window, int x, int y, int w, int h) {
		MoveWindow(window, x, y, w, h, true);
	}
	
	void Window::os::setTitle(Window window, const std::string &title) {
		SetWindowText(window, title.c_str());
	}
			
	void Window::os::show(Window window) {
		ShowWindow(window, SW_SHOWNORMAL);
	}
	
	void Window::os::hide(Window window) {
		ShowWindow(window, SW_HIDE);
	}

	void Window::os::zoom(Window window) {
		ShowWindow(window, SW_MAXIMIZE);
	}

	void Window::os::unzoom(Window window) {
		ShowWindow(window, SW_RESTORE);
	}

	void Window::os::showCursor() {
		ShowCursor(TRUE);
	}

	void Window::os::hideCursor() {
		ShowCursor(FALSE);
	}

	void Window::os::addHandler(hsm::Window *hw, EventHandler *eh) {
		hw->handlers.push_front(eh->getHandler());
	}
	
	void Window::os::removeHandler(hsm::Window *hw, EventHandler *eh) {
		hw->handlers.remove(eh->getHandler());
	}


/* loading action *************************************************************/

	namespace {
				
		class Loader {
		private :
			static const Loader instance;
		
		private :
			HWND consoleWindow;

		private :
			Loader() {
				char consoleTitle[255];
				GetConsoleTitle(consoleTitle, 255);
				consoleWindow = FindWindow("ConsoleWindowClass", consoleTitle);
				HSM_FINAL( ShowWindow(consoleWindow, SW_HIDE); )
			}

			~Loader() {
				HSM_FINAL( ShowWindow(consoleWindow, SW_SHOWNORMAL); )
			}
		};

		const Loader Loader::instance;


	} // namespace

} // namespace hsm
