////////////////////////////////////////////////////////////////////////////////
// File    : Tablet.MACOSX.cpp
// Content : libhsm system tablet plugin
// History : 22-Oct-2001 - [rb] creation
////////////////////////////////////////////////////////////////////////////////

/*
    Copyright (C) 2004  Renaud Blanch <rb.hsm@laposte.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// ToDo ////////////////////////////////////////////////////////////////////////
// - device types are only Wacom pen, eraser and 4DMouse
// - [BUG] no proximity is sent if device enter the tablet outside a window
// + convert tablet coordinates in "screen" coordinates (or let the system do
//   it ?)
 

/* includes *******************************************************************/

#include <Tablet.h>

#include <hsm/component/Point.h>
#include <hsm/component/ButtonPad.h>
#include <hsm/system/window/Window.h>
#include <hsm/util/Debug.h>

#include <string>
#include <sstream>
#include <map>


namespace wacom {
	
/* wacom helper functions *****************************************************/

	UInt32 contextId;
		
	OSErr	createWacomContextForTablet(UInt32 tabletIdx, UInt32 *contextID);
	OSErr	destroyWacomContext(UInt32	contextID);
	OSErr setContext(UInt32 contextID);

} // namespace wacom
	

namespace hsm {
	
/* Tablet instance ************************************************************/

	Tablet Tablet::instance;


/* Tablet implementation ******************************************************/

	enum DevType {
		STYLUS = 1,
		MOUSE4D,
		ERASER
	};

	std::map< unsigned int, unsigned int > devTypeMap;

	Tablet::Tablet() : impl::Component< EventHandler >("Tablet") {
		HSM_PRINT(1, "[Tablet::Tablet()]");

		nbEvents = 2;
		eventSpec = new Window::os::EventSpec[nbEvents];
		eventSpec[0].eventClass = kEventClassTablet;
		eventSpec[0].eventKind = kEventTabletPoint;
		eventSpec[1].eventClass = kEventClassTablet;
		eventSpec[1].eventKind = kEventTabletProximity;
		wacom::createWacomContextForTablet(1, &wacom::contextId);
		wacom::setContext(wacom::contextId);
		
		hsm::repository::add(this);
	}
	
	Tablet::~Tablet() {
		HSM_PRINT(1, "[Tablet::~Tablet()]");

		hsm::repository::remove(this);
		delete[] eventSpec;
		wacom::destroyWacomContext(wacom::contextId);
	}

	OSStatus Tablet::handler(EventHandlerCallRef inRef, EventRef inEvent, void *userData) {
		TabletComponent *device = 0;
		std::stringstream deviceId;
		unsigned int devId;
		unsigned int type;

		TabletPointRec     point;
		TabletProximityRec proximity;
		bool leaving;

		switch(GetEventKind(inEvent)) {
		case kEventTabletPoint :
			GetEventParameter(inEvent, kEventParamTabletPointRec, typeTabletPointRec, 0, sizeof(TabletPointRec), 0, &point);
			devId = point.deviceID;
			type = devTypeMap[devId];
			deviceId << "dev-" << devId << "-" << type << "-";
			device = dynamic_cast< TabletComponent * >(instance.components[deviceId.str()]);
			if(device == 0) {
				return eventNotHandledErr;
			}

			*(device->position) = point.absX, point.absY;			
			*(device->buttons) = point.buttons;

			switch(type) {
			case MOUSE4D :
				*(device->thumbwheel) = point.absZ;
				*(device->orientation) = point.rotation / 64.;
				break;

			case STYLUS :
			case ERASER :
				*(device->pressure) = point.pressure;
				*(device->orientation) = point.tiltX / 32767. * 90, point.tiltY / 32767. * 90;
				break;
			}
			break;

		case kEventTabletProximity :
			GetEventParameter(inEvent, kEventParamTabletProximityRec, typeTabletProximityRec, 0, sizeof(TabletProximityRec), 0, &proximity);
			if(proximity.pointerType == 0) {
				break;
			}
			
			// device id
			type = proximity.pointerType;
			devId = proximity.deviceID;
			deviceId << "dev-" << devId << "-" << type << "-";

			leaving = (proximity.enterProximity == 0);			
			

			// entering proximity
			if(!leaving) {
				devTypeMap[devId] = type;
				device = dynamic_cast<TabletComponent *>(instance.components[deviceId.str()]);

				// new device
				if(device == 0) {
					device = new TabletComponent(deviceId.str());

					device->add(device->position = new Point("position", 2));
					switch(type) {
					case MOUSE4D :
						device->add(device->thumbwheel = new Point("thumbwheel", 1));
						device->add(device->orientation = new Point("orientation", 1));
						device->add(device->buttons = new ButtonPad("buttons", 5));
						break;
					case STYLUS :
						device->add(device->pressure = new Point("pressure", 1));
						device->add(device->orientation = new Point("orientation", 2));
						device->add(device->buttons = new ButtonPad("buttons", 3));
						break;
					case ERASER :
						device->add(device->pressure = new Point("pressure", 1));
						device->add(device->orientation = new Point("orientation", 2));
						device->add(device->buttons = new ButtonPad("buttons", 1));
						break;
					}

					instance.add(device);
				}
			}

			break;

		} // switch(GetEventClass(inEvent))

		return eventNotHandledErr;
	}

	Window::os::Handler *Tablet::getHandler() const {
		return handler;
	}


} // namespace hsm


namespace wacom {

/* wacom helper functions *****************************************************/

	// taken from TabletAEDictionary included in tableteventdemo.sit
	// see [http://www.wacomeng.com/devsupport/mac/downloads.html]
		
	#define kDefaultTimeOut 15
		
	#define kWacomDriverSig		'WaCM'
	#define kAEWacomSuite		'Wacm'

	#define cContext				'CTxt'
	#define cWTDTablet			'Tblt'

	#define pMapScreenArea		'Smap'	// typeQDRectangle r/w

	#define pContextMovesSystemCursor	'Mvsc'				// typeBoolean
	#define pContextMapScreenArea			pMapScreenArea		// typeQDRectangle
	#define pContextMapTabletOutputArea	'Tomp'				// typeLongRectangle
	
	

	OSErr	createWacomContextForTablet(UInt32 tabletIdx, UInt32 *contextID) {
		OSErr err;

		/* open the driver */
		OSType tdSig = kWacomDriverSig;
		AEDesc driverTarget;
		
		err = AECreateDesc(typeApplSignature, (Ptr) &tdSig, sizeof(tdSig), &driverTarget);
		if(err) {
			AEDisposeDesc(&driverTarget);
			return err;
		}

		/* ask for a context */
		// creating the event
		AppleEvent aeSend;
		err = AECreateAppleEvent(kAECoreSuite, kAECreateElement, &driverTarget, kAutoGenerateReturnID, kAnyTransactionID, &aeSend);

		// filling the event
		DescType   contextDescType = cContext;
		err = AEPutParamPtr(&aeSend, keyAEObjectClass, typeType, &contextDescType, sizeof(DescType));
			
		AEDesc nullDesc, keyData, tObjSpecifier;
		AEInitializeDesc(&nullDesc);
		err = AECreateDesc(typeNull, 0, 0, &nullDesc);
		AEInitializeDesc(&keyData);
		err = AECreateDesc(typeUInt32, &tabletIdx, sizeof(tabletIdx), &keyData);
		err = CreateObjSpecifier(cWTDTablet, &nullDesc, formAbsolutePosition, &keyData, true, &tObjSpecifier);
		err = AEPutParamDesc(&aeSend, keyAEInsertHere,  &tObjSpecifier);

		// sending the event
		AppleEvent aeReply;
		err = AESend(&aeSend, &aeReply, kAEWaitReply, kAEHighPriority, kDefaultTimeOut, 0, 0);
			
		/* cleaning */
		AEDisposeDesc(&tObjSpecifier);
		AEDisposeDesc(&aeSend);

		/* getting the Context ID out of the reply */
		UInt32     outContextID;
		DescType   outType;
		Size       outSize;
		err = AEGetParamPtr(&aeReply, keyDirectObject, typeUInt32, &outType, &outContextID, sizeof(long), &outSize);
		if(!err) {
			if(contextID) *contextID = outContextID;		
		}

		AEDisposeDesc(&driverTarget);
		AEDisposeDesc(&aeReply);

		return err;
	}
		
		
	OSErr	destroyWacomContext(UInt32	contextID) {
		OSErr err;

		/* open the driver */
		OSType tdSig = kWacomDriverSig;
		AEDesc driverTarget;
		
		err = AECreateDesc(typeApplSignature, (Ptr) &tdSig, sizeof(tdSig), &driverTarget);
		if(err) {
			AEDisposeDesc(&driverTarget);
			return err;
		}

		/* destroying the context */
		// creating the event
		AppleEvent aeSend;
		err = AECreateAppleEvent(kAECoreSuite, kAECreateElement, &driverTarget, kAutoGenerateReturnID, kAnyTransactionID, &aeSend);

		// filling the event
		AEDesc nullDesc, keyData, tObjSpecifier;
		AEInitializeDesc(&nullDesc);
		err = AECreateDesc(typeNull, 0, 0, &nullDesc);

		AEInitializeDesc(&keyData);
		err = AECreateDesc(typeUInt32, &contextID, sizeof(contextID), &keyData);
		err = CreateObjSpecifier(cContext, &nullDesc, formUniqueID, &keyData, true, &tObjSpecifier);
		err = AEPutParamDesc(&aeSend, keyAEInsertHere,  &tObjSpecifier);

		// sending the event
		err = AESend(&aeSend, 0, kAEWaitReply, kAEHighPriority, kDefaultTimeOut, 0, 0);

		/* cleaning */
		AEDisposeDesc(&driverTarget);
		AEDisposeDesc(&tObjSpecifier);
		AEDisposeDesc(&aeSend);

		return err;
	}



	OSErr setContext(UInt32 contextID) {
		OSErr err;

		OSType tdSig = kWacomDriverSig;
		AEDesc driverTarget;

		AppleEvent aeSend, aeReply;

		DescType contextMapScreenAreaDescType = typeQDRectangle;
		Rect     outMapScreenArea;

		AEDesc nullDesc, keyData, tObjSpecifier, lastDesc;
			
		DescType attribute;
		DescType outType;
		Size     outSize;
			
		/* open the driver */
		err = AECreateDesc(typeApplSignature, (Ptr) &tdSig, sizeof(tdSig), &driverTarget);
		if(err) {
			AEDisposeDesc(&driverTarget);
			return err;
		}

		/* retrieving tablet and screen mapping */
		// creating the event
		err = AECreateAppleEvent(kAECoreSuite, kAEGetData, &driverTarget, kAutoGenerateReturnID, kAnyTransactionID, &aeSend);

		// filling the event
		AEInitializeDesc(&nullDesc);
		err = AECreateDesc(typeNull, 0, 0, &nullDesc);

		AEInitializeDesc(&keyData);
		err = AECreateDesc(typeUInt32, &contextID, sizeof(contextID), &keyData);
		err = CreateObjSpecifier(cContext, &nullDesc, formUniqueID, &keyData, true, &tObjSpecifier);
		err = AEPutParamDesc(&aeSend, keyAEInsertHere,  &tObjSpecifier);

		AEDuplicateDesc(&tObjSpecifier, &lastDesc);
		AEDisposeDesc(&tObjSpecifier);
	
		attribute = pContextMapScreenArea;
		AEInitializeDesc(&keyData);
		err = AECreateDesc(typeType, &attribute, sizeof(DescType), &keyData);
		err = CreateObjSpecifier(cProperty, &lastDesc, formPropertyID, &keyData, true, &tObjSpecifier);
		err = AEPutParamPtr(&aeSend, keyAERequestedType, typeType, &contextMapScreenAreaDescType, sizeof(DescType));
		err = AEPutParamDesc(&aeSend, keyDirectObject, &tObjSpecifier);

		// sending the event
		err = AESend(&aeSend, &aeReply, kAEWaitReply, kAEHighPriority, kDefaultTimeOut, 0, 0);

		// receiving the response
		err = AEGetParamPtr(&aeReply, keyDirectObject, contextMapScreenAreaDescType, &outType, &outMapScreenArea, sizeof(Rect), &outSize);

		// cleaning
		AEDisposeDesc(&tObjSpecifier);
		AEDisposeDesc(&aeSend);
		AEDisposeDesc(&aeReply);


		/* stop moving the cursor */
		bool moveCursor = false;

		// creating the event
		err = AECreateAppleEvent(kAECoreSuite, kAESetData, &driverTarget, kAutoGenerateReturnID, kAnyTransactionID, &aeSend);

		// filling the event
		AEInitializeDesc(&nullDesc);
		err = AECreateDesc(typeNull, 0, 0, &nullDesc);

		AEInitializeDesc(&keyData);
		err = AECreateDesc(typeUInt32, &contextID, sizeof(contextID), &keyData);
		err = CreateObjSpecifier(cContext, &nullDesc, formUniqueID, &keyData, true, &tObjSpecifier);
		err = AEPutParamDesc(&aeSend, keyAEInsertHere,  &tObjSpecifier);

		AEDuplicateDesc(&tObjSpecifier, &lastDesc);
		AEDisposeDesc(&tObjSpecifier);
	
		attribute = pContextMovesSystemCursor;
		AEInitializeDesc(&keyData);
		err = AECreateDesc(typeType, &attribute, sizeof(DescType), &keyData);
		err = CreateObjSpecifier(cProperty, &lastDesc, formPropertyID, &keyData, true, &tObjSpecifier);
		err = AEPutParamPtr(&aeSend, keyAEData, typeBoolean, &moveCursor, sizeof(bool));
		err = AEPutParamDesc(&aeSend, keyDirectObject, &tObjSpecifier);	

		// sending the event
		err = AESend(&aeSend, 0, kAEWaitReply, kAEHighPriority, kDefaultTimeOut, 0, 0);

		// cleaning
		AEDisposeDesc(&tObjSpecifier);
		AEDisposeDesc(&aeSend);


		/* setting the tablet output area to match screen coordinates */
		struct LongRect {
			long top;
			long left;
			long bottom;
			long right;
		};
		LongRect rect;
		rect.top = outMapScreenArea.top;
		rect.left = outMapScreenArea.left;
		rect.bottom = outMapScreenArea.bottom;
		rect.right = outMapScreenArea.right;
			
		// creating the event
		err = AECreateAppleEvent(kAECoreSuite, kAESetData, &driverTarget, kAutoGenerateReturnID, kAnyTransactionID, &aeSend);

		// filling the event
		AEInitializeDesc(&nullDesc);
		err = AECreateDesc(typeNull, 0, 0, &nullDesc);

		AEInitializeDesc(&keyData);
		err = AECreateDesc(typeUInt32, &contextID, sizeof(contextID), &keyData);
		err = CreateObjSpecifier(cContext, &nullDesc, formUniqueID, &keyData, true, &tObjSpecifier);
		err = AEPutParamDesc(&aeSend, keyAEInsertHere,  &tObjSpecifier);

		AEDuplicateDesc(&tObjSpecifier, &lastDesc);
		AEDisposeDesc(&tObjSpecifier);
	
		attribute = pContextMapTabletOutputArea;
		AEInitializeDesc(&keyData);
		err = AECreateDesc(typeType, &attribute, sizeof(DescType), &keyData);
		err = CreateObjSpecifier(cProperty, &lastDesc, formPropertyID, &keyData, true, &tObjSpecifier);
		err = AEPutParamPtr(&aeSend, keyAEData, typeLongRectangle, &rect, sizeof(rect));
		err = AEPutParamDesc(&aeSend, keyDirectObject, &tObjSpecifier);	

		// sending the event
		err = AESend(&aeSend, 0, kAEWaitReply, kAEHighPriority, kDefaultTimeOut, 0, 0);

		// cleaning
		AEDisposeDesc(&tObjSpecifier);
		AEDisposeDesc(&aeSend);


		/* cleaning */
		AEDisposeDesc(&driverTarget);

		return err;
	}
	
} // namespace wacom
